/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DatabaseImpl.java 564998 2007-08-12 02:51:49Z vgritsenko $
 */

package org.apache.xindice.client.xmldb.embed;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.client.xmldb.CommonConfigurable;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.Database;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.server.Xindice;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.ReadOnlyException;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.util.XindiceRuntimeException;
import org.apache.xindice.xml.dom.DOMParser;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * Implements XML:DB's <code>Database</code> interface providing embedded access to
 * a Xindice database. Usually, this class is not used
 * directly, but {@link org.apache.xindice.client.xmldb.DatabaseImpl} is used instead.
 *
 * Embedded database driver supports multiple database instances simultaneously. To use
 * multiple instances, configure several <code>root-collection</code> elements in the
 * Xindice configuration file and point this driver to this configuration.
 *
 * Currently, databases are loaded by the driver lazily, on first access to the
 * database.
 *
 * Embedded database driver uses following configuration parameters:
 * <ul>
 * <li><strong>db-home</strong>: Specifies path to the Xindice DB Home directory</li>
 * <li><strong>configuration</strong>: Specifies path to the Xindice Configuration file</li>
 * <li><strong>managed</strong>: If set to true, driver will not load databases by himself,
 * but will assume that third party code had initialized the database
 * (See {@link org.apache.xindice.client.xmldb.managed.DatabaseImpl})</li>
 * </ul>
 *
 * If one of these parameters is not specified, system properties will be used.
 *
 * Driver parameters can be specified either:
 * <ul>
 * <li>by instantiating {@link org.apache.xindice.client.xmldb.DatabaseImpl}
 *     and setting its properties (see {@link #setProperty}). You won't use this class
 *     directly in this case.</li>
 * <li>by setting parameters via {@link #setProperty} method <strong>before</strong>
 *     obtaining collection with {@link #getCollection(String, String, String)} method.</li>
 * <li>by setting Java system properties (see {@link System#setProperty(String, String)})</li>
 * <li>by using constructor {@link DatabaseImpl#DatabaseImpl(CommonConfigurable)} and
 *     passing instance of CommonConfigurable with all properties already set.</li>
 *
 * @author <a href="mailto:kstaken@xmldatabases.org">Kimbro Staken</a>
 * @author <a href="mailto:james.bates@amplexor.com">James Bates</a>
 * @author <a href="mailto:vladimir@apache.org">Vladimir R. Bossicard</a>
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 * @version $Revision: 564998 $, $Date: 2007-08-11 22:51:49 -0400 (Sat, 11 Aug 2007) $
 */
public class DatabaseImpl extends CommonConfigurable implements org.xmldb.api.base.Database {

    private static final Log log = LogFactory.getLog(DatabaseImpl.class);

    /**
     * Driver property name for the xindice database home.
     */
    public static final String PROP_XINDICE_DB_HOME = "db-home";

    /**
     * System property name for the xindice database home.
     */
    public static final String SYSPROP_XINDICE_DB_HOME = Xindice.PROP_XINDICE_DB_HOME;

    /**
     * Driver property name for the xindice configuration file.
     */
    public static final String PROP_XINDICE_CONFIGURATION = "configuration";

    /**
     * System property name for the xindice configuration file.
     */
    public static final String SYSPROP_XINDICE_CONFIGURATION = Xindice.PROP_XINDICE_CONFIGURATION;

    /**
     * Driver property name for the managed configuration parameter.
     */
    public static final String PROP_XINDICE_MANAGED = "db-managed";

    /**
     * System property name for the managed configuration parameter.
     */
    public static final String SYSPROP_XINDICE_MANAGED = "xindice.db.managed";

    /**
     * Prefix used to denote XML:DB URI's that should use this driver
     */
    public static final String DRIVER_NAME = "xindice-embed";

    /**
     * XML:DB conformance level of this driver
     */
    private static final String CONFORMANCE_LEVEL = "0";


    /**
     * Database configuration location.
     * Null if built-in configuration was used.
     */
    private String configFile;


    /**
     * Creates new <code>DatabaseImpl</code> instance.
     */
    public DatabaseImpl() {
    }

    /**
     * Create a new DatabaseImpl object with a copy of the properties
     * from the <code>config</code> parameter.
     *
     * This allows to pass properties such as PROP_XINDICE_CONFIGURATION
     * to this instance. Usually this is done by instantiating not this
     * class, but {@link org.apache.xindice.client.xmldb.DatabaseImpl},
     * set all the necessary parameters, and then get a collection.
     *
     * @param config from which the initial parameters for this
     *        DatabaseImpl object are copied.
     */
    public DatabaseImpl(CommonConfigurable config) {
        super(config);
    }

    /**
     * Get named database instance. If database is not loaded yet, load it.
     *
     * @param dbName Database name
     * @param uri Database URI
     * @return Database instance
     * @throws XMLDBException if configuration for this database is not found,
     *                        or could not be read.
     */
    private Database getDatabase(String dbName, String uri) throws XMLDBException {
        Database database = Database.getDatabase(dbName);
        if (database == null && !isManaged()) {
            Configuration dbConfig = getConfiguration(dbName);
            if (log.isDebugEnabled()) {
                log.debug("Mounting database: '" + dbName + "'");
            }

            if (dbConfig == null) {
                throw new XMLDBException(ErrorCodes.NO_SUCH_DATABASE,
                                         "Database '" + dbName + "' not found: " + uri);
            }

            try {
                database = Database.getDatabase(dbConfig);
            } catch (DBException e) {
                throw FaultCodes.createXMLDBException(e);
            }
            if (log.isDebugEnabled()) {
                log.info("Mounted database: '" + database.getName() + "'");
            }
        }

        return database;
    }

    protected boolean isManaged() {
        String managed = null;
        try {
            // Try configuration first
            managed = getProperty(PROP_XINDICE_MANAGED);
        } catch (XMLDBException e) {
            /* ignored */
        }
        if (managed == null) {
            // Fallback to system property
            managed = System.getProperty(SYSPROP_XINDICE_MANAGED);
        }
        return Boolean.valueOf(managed).booleanValue();
    }

    /**
     * Get configuration for the named database
     *
     * @param dbName Database name
     * @return Database configuration
     * @throws XMLDBException if unable to load database configuration
     */
    protected Configuration getConfiguration(String dbName) throws XMLDBException {
        Configuration config = loadConfiguration();

        // Find database config with the given database name
        Configuration[] roots = config.getChildren("root-collection");
        config = null;
        for (int i = 0; i < roots.length; i++) {
            if (dbName.equals(roots[i].getAttribute(Database.NAME))) {
                config = roots[i];
                break;
            }
        }

        if (config == null) {
            return null;
        }

        // Figure out database root
        String dbRoot = config.getAttribute(Database.DBROOT, Database.DBROOT_DEFAULT);
        if (!new File(dbRoot).isAbsolute()) {
            // Let's see if the XINDICE_DB_HOME property was specified.
            String home = null;
            try {
                // Try configuration first
                home = getProperty(PROP_XINDICE_DB_HOME);
            } catch (XMLDBException e) {
                /* ignored */
            }
            if (home == null) {
                // Fallback to system property
                home = System.getProperty(SYSPROP_XINDICE_DB_HOME);
            }

            if (home != null) {
                try {
                    dbRoot = new File(home + File.separator + dbRoot).getCanonicalPath();
                } catch (IOException e) {
                    log.warn("getCanonicalPath failed", e);
                    dbRoot = new File(home + File.separator + dbRoot).getAbsolutePath();
                }
            } else if (configFile != null) {
                dbRoot = new File(configFile).getAbsoluteFile().getParent() + File.separator + dbRoot;
            } else {
                log.warn("The database configuration file is not specified and there was no "
                         + SYSPROP_XINDICE_DB_HOME + " property set, "
                         + "so Xindice was unable to determine a database location. "
                         + "Database will be created relative to the current directory.");
                try {
                    dbRoot = new File("." + File.separator + dbRoot).getCanonicalPath();
                } catch (IOException e) {
                    log.warn("getCanonicalPath failed", e);
                    dbRoot = new File("." + File.separator + dbRoot).getAbsolutePath();
                }
            }
            try {
                config.setAttribute(Database.DBROOT, dbRoot);
            } catch (ReadOnlyException e) {
                /* ignored */
            }
        }

        return config;
    }

    private Configuration loadConfiguration() throws XMLDBException {
        Configuration config;

        if (configFile == null) {
            try {
                // Try configuration first
                configFile = getProperty(PROP_XINDICE_CONFIGURATION);
            } catch (XMLDBException e) {
                /* ignored */
            }
            if (configFile == null) {
                // Fallback to system property
                configFile = System.getProperty(SYSPROP_XINDICE_CONFIGURATION);
            }
            if ("".equals(configFile)) {
                configFile = null;
            }
        }

        if (configFile != null) {
            if (log.isInfoEnabled()) {
                log.info("Specified configuration file: '" + configFile + "'");
            }
            try {
                FileInputStream configXMLFile = new FileInputStream(configFile);
                config = new Configuration(DOMParser.toDocument(configXMLFile), false);
            } catch (XindiceException e) {
                throw new XMLDBException(ErrorCodes.UNKNOWN_ERROR,
                                         "Unable to parse database configuration file: " + configFile, e);
            } catch (FileNotFoundException e) {
                throw new XMLDBException(ErrorCodes.UNKNOWN_ERROR,
                                         "Database configuration file not found: " + configFile);
            }
        } else {
            if (log.isInfoEnabled()) {
                log.info("No configuration file specified, going with the default configuration");
            }

            try {
                config = new Configuration(DOMParser.toDocument(Xindice.DEFAULT_CONFIGURATION), false);
            } catch (XindiceException e) {
                throw new XindiceRuntimeException("Unable to parse default database configuration", e);
            }
        }

        return config;
    }

    /**
     * Determines whether this <code>Database</code> implementation  can handle
     * the URI. It should return true if the Database instance knows how to
     * handle the URI and false otherwise.
     *
     * @param uri the URI to check for.
     * @return true if the URI can be handled, false otherwise.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     */
    public boolean acceptsURI(String uri) throws XMLDBException {
        return ((uri != null) && uri.startsWith(getName() + "://"));
    }

    /**
     * Retrieves a <code>Collection</code> instance based on the URI provided
     * in the <code>uri</code> parameter. The format of the URI is defined in the
     * documentation for DatabaseManager.getCollection().<p/>
     *
     * Authentication is handled via username and password however it is not
     * required that the database support authentication. Databases that do not
     * support authentication MUST ignore the
     * <code>username</code> and <code>password</code> if those provided are not
     * null.
     *
     * @param uri the URI to use to locate the collection.
     * @param password The password to use for authentication to the database or
     *    null if the database does not support authentication.
     * @return A <code>Collection</code> instance for the requested collection or
     *  null if the collection could not be found.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     *  <code>ErrroCodes.PERMISSION_DENIED</code> If the <code>username</code>
     *    and <code>password</code> were not accepted by the database.
     */
    public Collection getCollection(String uri, String userName, String password) throws XMLDBException {
        /* TODO: introduce authentication some day */

        if (!acceptsURI(uri)) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL: " + uri);
        }

        /* Chop off driver prefix, and '://' */
        uri = uri.substring(getName().length() + 3);

        /* Extract host name & port, if present */
        int firstSlash = uri.indexOf('/');
        if (firstSlash == -1) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (must have '/'): " + uri);
        }

        /* Extract collection name */
        String collPath = uri.substring(firstSlash);
        if (!collPath.startsWith("/")) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (collection name must start with '/'): " + uri);
        }

        // Find the database name. We just skip the first slash
        int colIndex = collPath.indexOf('/', 1);

        // We assume there's no collection specified
        String dbName = collPath.substring(1);

        // If colIndex isn't -1 then we need to pick out the db name
        if (colIndex != -1) {
            dbName = collPath.substring(1, colIndex);
        }

        Database database = getDatabase(dbName, uri);
        if (database == null) {
            throw new XMLDBException(ErrorCodes.NO_SUCH_DATABASE,
                                     "Database '" + dbName + "' not found. URL: " + uri);
        }

        try {
            return new CollectionImpl(database, collPath);
        } catch (XMLDBException e) {
            if (e.errorCode == ErrorCodes.NO_SUCH_COLLECTION) {
                // Per getCollection contract, return null if not found
                return null;
            }
            throw e;
        }
    }

    /**
     * Returns the prefix used in XML:DB to denote URI's that this driver can
     * handle.
     *
     * @return the prefix driver name
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getName() throws XMLDBException {
        return DRIVER_NAME;
    }

    /**
     * Returns an array of names associated with the Database instance.
     *
     * @return the array of name of the object.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String[] getNames() throws XMLDBException {
        return new String[]{ getName() };
    }

    /**
     * Returns the XML:DB API Conformance level for the implementation. This can
     * be used by client programs to determine what functionality is available to
     * them.
     *
     * @return the XML:DB API conformance level for this implementation.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getConformanceLevel() throws XMLDBException {
        return CONFORMANCE_LEVEL;
    }
}
