/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: CollectionManagementServiceImpl.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.client.xmldb.services;

import org.apache.xindice.client.xmldb.CommonConfigurable;
import org.apache.xindice.client.xmldb.XindiceCollection;
import org.apache.xindice.core.FaultCodes;

import org.w3c.dom.Document;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.modules.CollectionManagementService;

/**
 * XML:DB CollectionManagementService implementation using XML-RPC to
 * communicate with Xindice.
 *
 * @author <a href="mailto:james.bates@amplexor.com">James Bates</a>
 * @author <a href="mailto:kstaken@xmldatabases.org">Kimbro Staken</a>
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class CollectionManagementServiceImpl extends CommonConfigurable
        implements CollectionManagementService, CollectionManager, DatabaseInstanceManager {

    /* The collection this service operates on */
    private Collection coll = null;

    /* service version */
    private final static String SERVICE_VERSION = "1.0";

    /* service name */
    private final static String SERVICE_NAME = "CollectionManagementService";

    /**
     * Returns service version
     *
     * @return <code>1.0</code>
     */
    public String getVersion() throws XMLDBException {

        return SERVICE_VERSION;
    }

    /**
     * Inserts new child collection into this collection
     *
     * @param childCollName name for new child collection
     * @return the newly created child
     */
    public Collection createCollection(String childCollName)
            throws XMLDBException {

        if (coll == null) {

            throw new XMLDBException(ErrorCodes.INVALID_COLLECTION,
                                     "Must set collection for this service");
        }

        return ((XindiceCollection) coll).createCollection(childCollName);
    }

    /**
     * Removes child collection from this collection
     *
     * @param childCollName name of child collection to remove
     */
    public void removeCollection(String childCollName) throws XMLDBException {
        if (coll == null) {

            throw new XMLDBException(ErrorCodes.INVALID_COLLECTION,
                                     "Must set collection for this service");
        }

        ((XindiceCollection) coll).removeCollection(childCollName);
    }

    /**
     * Returns service name
     *
     * @return <code>CollectionManagementService</code>
     */
    public String getName() {

        return SERVICE_NAME;
    }

    /**
     * Sets collection this service should operate on
     *
     * @param collection the collection this service should operate on
     */
    public void setCollection(Collection collection) {

        this.coll = collection;
    }

    /**
     * Returns the name of the collection that this manager is associated with.
     *
     * @return the name of the collection
     * @exception XMLDBException
     */
    public String getCollectionName() throws XMLDBException {
        try {
            return coll.getName();
        } catch (Exception e) {
            throw FaultCodes.createXMLDBException(e);
        }
    }

    /**
     * Returns the fully qualified name of the collection that this manager is
     * associated with. This name includes all parent collections.
     *
     * @return the fully qualified name for this collection.
     * @exception XMLDBException
     */
    public String getCanonicalName() throws XMLDBException {
        try {
            return ((XindiceCollection) coll).getCanonicalName();
        } catch (Exception e) {
            throw FaultCodes.createXMLDBException(e);
        }
    }

    /**
     * Creates a new collection in the database identified by name and using
     * the provided configuration.
     *
     * @param path the path of the new collection
     * @param configuration the XML collection configuration to use for
     *        creating this collection.
     * @return The newly created collection
     * @exception XMLDBException
     */
    public Collection createCollection(String path, Document configuration)
            throws XMLDBException {
        if (coll == null) {
            throw new XMLDBException(ErrorCodes.INVALID_COLLECTION,
                                     "Must set collection for this service");
        }

        return ((XindiceCollection) coll).createCollection(path, configuration);
    }

    /**
     * Drops a child collection from this collection.
     *
     * @param name The name of the child collection to drop.
     * @exception XMLDBException
     */
    public void dropCollection(String name) throws XMLDBException {
        this.removeCollection(name);
    }

    /**
     * Returns a list of all indexers for this collection.
     *
     * @return the list of indexers
     * @exception XMLDBException
     */
    public String[] listIndexers() throws XMLDBException {
        try {
            return ((XindiceCollection) coll).listIndexers();
        } catch (Exception e) {
            throw FaultCodes.createXMLDBException(e);
        }
    }

    /**
     * Creates a new Indexer for this collection.
     *
     * @param configuration The configuration to use for this indexer.
     * @exception XMLDBException
     */
    public void createIndexer(Document configuration) throws XMLDBException {
        try {
            ((XindiceCollection) coll).createIndexer(configuration);
        } catch (Exception e) {
            throw FaultCodes.createXMLDBException(e);
        }
    }

    /* see superclass for documentation */
    public void dropIndexer(String name) throws XMLDBException {
        try {
            ((XindiceCollection) coll).dropIndexer(name);
        } catch (Exception e) {
            throw FaultCodes.createXMLDBException(e);
        }
    }

    /**
     * Shutsdown the Database instance
     *
     * @exception XMLDBException
     */
    public void shutdown() throws XMLDBException {
        ((XindiceCollection) coll).shutdown();

    }


    /**
     * Returns a list of all collection level XMLObjects for this collection.
     *
     * @return the list of XMLObjects.
     * @exception XMLDBException
     */
    public String[] listXMLObjects() throws XMLDBException {
        return null;
    }

    /**
     * Creates a new collection level XMLObject using the provided configuration.
     * The XMLObject will be added to the collection using the provided name as
     * a key.
     *
     * @param configuration The XML configuration to use
     * @exception XMLDBException
     */
    public void createXMLObject(Document configuration) throws XMLDBException {

    }

    /**
     * Drops a collection level XMLObject from the collection.
     *
     * @param name The name of the XMLObject to drop.
     * @exception XMLDBException
     */
    public void dropXMLObject(String name)
            throws XMLDBException {

    }
}
