/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SystemCollection.java 568786 2007-08-23 00:40:57Z vgritsenko $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.SymbolTable;
import org.apache.xindice.xml.SymbolTableSymbols;
import org.apache.xindice.xml.dom.DOMParser;

import org.w3c.dom.Document;

/**
 * SystemCollection represents the System Collection.  Beyond
 * standard Collection operations, this class will provide facilities
 * for Schema and SymbolTable management among other things.
 *
 * @version $Revision: 568786 $, $Date: 2007-08-22 20:40:57 -0400 (Wed, 22 Aug 2007) $
 */
public final class SystemCollection extends Collection {

    private static final Log log = LogFactory.getLog(SystemCollection.class);

    public static final String SYSCOL = "system";
    public static final String SYMBOLS = "SysSymbols";
    public static final String OBJECTS = "SysObjects";
    public static final String CONFIGS = "SysConfig";
    public static final String USERS = "SysUsers";
    public static final String GROUPS = "SysGroups";
    public static final String ACCESS = "SysAccess";

    private static String SYSCOL_DEFINITION
            = "<collection name='" + SYSCOL + "'>"
            // System Collections
            + "   <collections>"
            //       Symbol Tables Collection
            + "      <collection name='" + SYMBOLS + "' compressed='true'>"
            + "         <filer class='org.apache.xindice.core.filer.BTreeFiler'/>"
            //          Textual Representation of Hard-coded Symbol Table
            +           SymbolTableSymbols.getDefinition()
            + "         <indexes/>"
            + "      </collection>"
            //       System Configuration Collection
            + "      <collection name='" + CONFIGS + "' compressed='false'>"
            + "         <filer class='org.apache.xindice.core.filer.BTreeFiler' pagecount='16'/>"
            + "         <indexes/>"
            + "      </collection>"
            + "   </collections>"
            + "</collection>";

    /**
     * @param db Its database
     */
    public SystemCollection(Database db) {
        super(db);
    }

    void init() throws DBException {
        // Bootstrap the System Collection
        try {
            Document sysDoc = DOMParser.toDocument(SYSCOL_DEFINITION);
            Configuration sysCfg = new Configuration(sysDoc, false);
            setConfig(sysCfg);
        } catch (Exception e) {
            log.fatal("Failed to generate system collection '" + SYSCOL + "'", e);

            // TODO: System.exit() CANNOT be good here.
            System.exit(1);
        }
    }

    private String getSymbolTableName(Collection col) {
        String name = col.getCanonicalName();
        // Strip database name
        int idx = name.indexOf('/', 1);
        name = name.substring(idx + 1);

        return name.replace('/', '_');
    }

    /**
     * loadSymbols retrieves the SymbolTable for the specified Collection.
     *
     * @param collection The Collection whose SymbolTable is required
     * @return The requested SymbolTable
     * @throws DBException if unable to load symbols
     */
    public SymbolTable loadSymbols(Collection collection) throws DBException {
        String name = getSymbolTableName(collection);

        Collection symCol = getCollection(SYMBOLS);

        SymbolTable symbols = (SymbolTable) symCol.getObject(name);
        if (symbols == null) {
            symbols = new SymbolTable();
        }
        return symbols;
    }

    /**
     * saveSymbols save the SymbolTable for the specified Collection.
     *
     * @param collection The Collection that owns the SymbolTable
     * @param symbols The SymbolTable
     * @throws DBException if unable to save symbols
     */
    public void saveSymbols(Collection collection, SymbolTable symbols) throws DBException {
        if (symbols != null) {
            String name = getSymbolTableName(collection);
            getCollection(SYMBOLS).setObject(name, symbols);
            symbols.resetDirty();
        }
    }

    /**
     * dropSymbols drop the SymbolTable for the specified Collection.
     *
     * @param collection The Collection that owns the SymbolTable
     * @throws DBException if unable to drop symbols
     */
    public void dropSymbols(Collection collection) throws DBException {
        String name = getSymbolTableName(collection);

        Collection symCol = getCollection(SYMBOLS);
        try {
            symCol.remove(name);
        } catch (DBException e) {
            if (e.faultCode == FaultCodes.COL_DOCUMENT_NOT_FOUND) {
                // ignore
                return;
            }
            throw e;
        }
    }

    public static String getDefinition() {
        return SYSCOL_DEFINITION;
    }
}
