/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Filer.java 541516 2007-05-25 02:46:51Z vgritsenko $
 */

package org.apache.xindice.core.filer;

import org.apache.xindice.core.DBException;
import org.apache.xindice.core.DBObject;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.Configurable;
import org.apache.xindice.util.Named;

import java.io.File;

/**
 * Filer is the low-level file management interface for Xindice.  A Filer object
 * is implemented in order to provide a data source to the Xindice Collection
 * class.  Filers are developed to perform transparent storage and retrieval to
 * and from heterogenous data sources (such as FTP, HTTP, RDBMS, etc...)
 *
 * @version $Revision: 541516 $, $Date: 2007-05-24 22:46:51 -0400 (Thu, 24 May 2007) $
 */
public interface Filer extends Named, DBObject, Configurable {

    /**
     * setLocation tells the Filer where to store its data.
     *
     * @param root The root under which to store data
     * @param location The name to use for storing data
     */
    void setLocation(File root, String location);

    /**
     * readRecord returns a Record from the Filer based on the specified
     * Key.
     *
     * @param key The Record's Key
     * @return The returned Record
     */
    Record readRecord(Key key) throws DBException;

    /**
     * readRecord returns a Record from the Filer based on the specified
     * Key containing filer meta information and value.
     *
     * @param key The Record's Key
     * @param metaOnly if true, resulting record contains only meta information
     * @return The returned Record
     */
    Record readRecord(Key key, boolean metaOnly) throws DBException;

    /**
     * writeRecord writes a Value to the Filer based on the specified Key.
     *
     * @param key The Record's Key
     * @param value The Record's Value
     * @return Written Record
     */
    Record writeRecord(Key key, Value value) throws DBException;

    /**
     * deleteRecord removes a Record from the Filer based on the
     * specified Key.
     *
     * @param key The Record's Key
     * @return Whether or not the Record was deleted
     */
    boolean deleteRecord(Key key) throws DBException;

    /**
     * getRecordCount returns the number of Records in the Filer.
     *
     * @return The Record count
     */
    long getRecordCount() throws DBException;

    /**
     * getRecordSet returns a RecordSet object for the current Filer.
     *
     * @return The Filer Enumerator
     */
    RecordSet getRecordSet() throws DBException;

    /**
     * flush forcefully flushes any unwritten buffers to disk.
     */
    void flush() throws DBException;
}
