/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SizeableMemFiler.java 541516 2007-05-25 02:46:51Z vgritsenko $
 */

package org.apache.xindice.core.filer;

import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.SimpleConfigurable;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * SizeableMemFiler is an In-Memory Filer implementation for Xindice.  SizeableMemFiler can be
 * used for temporary collections and caching.  It's basically a layering on
 * top of HashMap.
 *
 * @author Terry Rosenbaum (terry@amicas.com)
 * @version $Revision: 541516 $, $Date: 2007-05-24 22:46:51 -0400 (Thu, 24 May 2007) $
 */
public final class SizeableMemFiler extends SimpleConfigurable implements Filer {

    public static final String INITIAL_SIZE_KEY = "org.apache.xindice.core.filer.SizeableMemFiler.INITIAL_SIZE";
    public static final int INITIAL_SIZE = 89;

    private Map hashTable = null;
    private boolean opened = false;
    private boolean readOnly = false;


    public SizeableMemFiler() {
        String aSizeString = System.getProperty(INITIAL_SIZE_KEY);
        int aSize = INITIAL_SIZE;
        try {
            if (aSizeString != null) {
                aSize = Integer.parseInt(aSizeString);
            }
        } catch (Exception anException) {
            aSize = INITIAL_SIZE;
        }
        hashTable = Collections.synchronizedMap(new HashMap(aSize));
    }

    public SizeableMemFiler(Map hashTable, boolean readOnly) {
        this.hashTable = hashTable;
        this.readOnly = readOnly;
    }

    public SizeableMemFiler(Map hashTable) {
        this(hashTable, false);
    }

    public void setLocation(File root, String location) {
    }

    public String getName() {
        return "SizeableMemFiler";
    }

    private void checkOpened() throws DBException {
        if (!opened) {
            throw new FilerException(FaultCodes.COL_COLLECTION_CLOSED, "Filer is closed");
        }
    }

    private void checkReadOnly() throws DBException {
        if (readOnly) {
            throw new FilerException(FaultCodes.COL_COLLECTION_READ_ONLY, "Filer is read-only");
        }
    }

    public boolean create() {
        hashTable.clear();
        return true;
    }

    public boolean open() {
        opened = true;
        return opened;
    }

    public boolean isOpened() {
        return opened;
    }

    public boolean exists() {
        return true;
    }

    public boolean drop() {
        hashTable.clear();
        opened = false;
        return !opened;
    }

    public boolean close() {
        opened = false;
        return !opened;
    }

    public void flush() {
    }

    public Record readRecord(Key key) throws DBException {
        return readRecord(key, false);
    }

    public Record readRecord(Key key, boolean metaOnly) throws DBException {
        if (key == null || key.getLength() == 0) {
            return null;
        }
        checkOpened();
        return (Record) hashTable.get(key);
    }

    public Record writeRecord(Key key, Value value) throws DBException {
        if (key == null || key.getLength() == 0) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid key: '" + key + "'");
        }
        if (value == null) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid null value");
        }
        checkOpened();
        checkReadOnly();
        hashTable.put(key, new Record(key, value));
        return new Record(key, value);
    }

    public boolean deleteRecord(Key key) throws DBException {
        if (key == null || key.getLength() == 0) {
            return false;
        }
        checkOpened();
        checkReadOnly();
        return hashTable.remove(key) != null;
    }

    public long getRecordCount() throws DBException {
        checkOpened();
        return hashTable.size();
    }

    public RecordSet getRecordSet() throws DBException {
        checkOpened();
        return new MemRecordSet();
    }

    /**
     * MemRecordSet
     */

    private class MemRecordSet implements RecordSet {
        private Iterator iter = hashTable.values().iterator();

        public synchronized boolean hasMoreRecords() throws DBException {
            return iter.hasNext();
        }

        public synchronized Record getNextRecord() throws DBException {
            checkOpened();
            return (Record) iter.next();
        }

        public synchronized Value getNextValue() throws DBException {
            checkOpened();
            return ((Record) iter.next()).getValue();
        }

        public synchronized Key getNextKey() {
            return ((Record) iter.next()).getKey();
        }
    }
}
