/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Searcher.java 584476 2007-10-14 02:41:50Z natalia $
 */

package org.apache.xindice.core.query.ftsearch;

import org.w3c.dom.NodeList;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.Query;
import org.apache.xindice.core.data.NodeSet;

import java.io.IOException;

/**
 * Searcher executes full text queries against list of nodes and returns
 * nodes that match the query in the order of relevance (score) - most relevant
 * results will be return first.<br>
 * <br>
 * Query syntax is the same as syntax of Lucene query, except it does not use
 * field names.
 *
 * @version $Revision: 584476 $, $Date: 2007-10-13 22:41:50 -0400 (Sat, 13 Oct 2007) $
 */
public class Searcher {
    private NodeList nodes;
    private Analyzer analyzer;

    /**
     * Builds new Searcher based on list of nodes and analyzer.
     *
     * @param nodes List of nodes to search
     * @param analyzer Analyzer that will be used to tokenize text of the nodes.
     * Choice of analyzer affects query results.
     * @see org.apache.lucene.analysis.Analyzer
     */
    public Searcher(NodeList nodes, Analyzer analyzer) {
        this.nodes = nodes;
        this.analyzer = analyzer;
    }

    /**
     * Executes query against list of nodes and returns matches in the order
     * of relevance (score).
     *
     * @param query Full text query
     * @return NodeSet that contains matching nodes
     * @throws ParseException Query failed to be parsed
     */
    public NodeSet search(String query) throws ParseException {
        Query compQuery = new QueryParser("", analyzer).parse(query);
        NodeReader reader = new NodeReader(nodes, analyzer);
        IndexSearcher searcher = new IndexSearcher(reader);
        Hits hits = null;
        try {
            hits = searcher.search(compQuery);
        } catch (IOException e) {
            // this searcher does not use file IO, exception won't happen
        }

        return new ResultSet(hits);
    }

    private class ResultSet implements NodeSet {
        private Hits hits;
        private int count;

        private ResultSet(Hits hits) {
            this.hits = hits;
        }

        public boolean hasMoreNodes() {
            return count < hits.length();
        }

        public Object getNextNode() {
            try {
                return nodes.item(hits.id(count++));
            } catch (IOException e) {
                // ignore, does not use IO
            }

            return null;
        }
    }
}
