/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: AddMultipleDocuments.java 578602 2007-09-23 20:33:31Z natalia $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.tools.XMLTools;

import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;

import java.io.File;
import java.io.FilenameFilter;

/**
 * AddMultipleDocuments.java is designed to let the user add several Documents
 * to an existing Collection at one time.
 *
 * Documents are added with an ID the same as the file name.
 * This command assumes that the collection being added to already exists!
 *
 * @version $Revision: 578602 $, $Date: 2007-09-23 16:33:31 -0400 (Sun, 23 Sep 2007) $
 */
public class AddMultipleDocuments extends Command {

    public boolean execute(XMLTools.Config table) throws Exception {

        // Verify that user has supplied necessary arguments
        if (table.getString(XMLTools.COLLECTION) == null) {
            System.out.println("ERROR : Collection and switch required");
            return false;
        }

        if ("".equals(table.getString(XMLTools.FILE_PATH))) {
            System.out.println("ERROR : Directory name and switch required");
            return false;
        }

        Collection col = null;
        try {

            // Get a Collection reference to the collection
            String colstring = normalizeCollectionURI(table.getString(XMLTools.COLLECTION),
                                                      table.getBoolean(XMLTools.LOCAL));
            col = DatabaseManager.getCollection(colstring);
            if (col == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            // Get a File object for the Directory passed in
            File dir = new File(table.getString(XMLTools.FILE_PATH));
            if (dir.isDirectory()) {

                String[] children = new String[]{};
                final String ext = table.getString(XMLTools.EXTENSION);

                // If the user supplied a file extension, filter on it, else use entire contents of the directory
                if (!ext.equals("")) {
                    children = dir.list(new FilenameFilter() {
                        public boolean accept(File none, String name) {
                            return name.endsWith("." + ext);
                        }
                    });
                } else if (ext.equals("")) {
                    children = dir.list();
                }

                System.out.println("Reading files from: " + dir.getCanonicalPath());

                // Use the functionality already provided in AddDocument command to add this document
                final Command cmd = new AddDocument();

                // Command parameters
                XMLTools.Config localtable = new XMLTools.Config();
                localtable.setString(XMLTools.COLLECTION, table.getString(XMLTools.COLLECTION));
                if (table.getBoolean(XMLTools.LOCAL)) {
                    localtable.setBoolean(XMLTools.LOCAL, true);
                }

                // Loop over documents, adding to db
                for (int i = 0; i < children.length; i++) {

                    File file = new File(dir, children[i]);
                    if (file.isFile()) {
                        try {
                            // Populate hashtable to pass to AddDocument class
                            localtable.setString(XMLTools.NAME_OF, file.getName());
                            localtable.setString(XMLTools.FILE_PATH, file.getPath());

                            // Execute the class!
                            cmd.execute(localtable);
                        } catch (Exception e) {
                            System.out.println("Error Adding File: " + file.getName());
                            if (table.getBoolean(XMLTools.VERBOSE)) {
                                e.printStackTrace();
                            }
                        }
                    }
                } // for loop
            }

        } finally {
            // Release all resources
            if (col != null) {
                col.close();
            }
        }

        return true;
    }

    public void usage() {
        System.out.println("Format: xindice addmultiple -c <context> [-l [-d <path>]] [-v] [parameters...]");
        System.out.println();
        System.out.println("Add several documents to an existing collection at one time. Documents are");
        System.out.println("added with an ID the same as the file name");
        System.out.println();
        System.out.println("Command-specific switches:");
        System.out.println("    -f|--filepath <path>");
        System.out.println("                 Directory name for the files being added to a collection");
        System.out.println("    -e|--extension <extention>");
        System.out.println("                 File extension for documents to use in the filter");
        System.out.println();
    }
}
