/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: FileCache.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Map;
import java.util.WeakHashMap;

/**
 * FileCache caches the content of files in memory.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class FileCache {

    /**
     * Caches FileCacheInfo objects. Keys are File objects.
     */
    private final Map cache = new WeakHashMap();

    public FileCache() {
    }

    public final boolean isInCache(File file) {
        return (cache.get(file) != null);
    }

    public final boolean isInCache(String name) {
        return (cache.get(new File(name)) != null);
    }

    public final boolean isModified(String name) {
        return isModified(new File(name));
    }

    public final boolean isModified(File file) {
        FileCacheInfo finfo = (FileCacheInfo) cache.get(file);
        return !file.exists()
                || finfo == null
                || (file.lastModified() != finfo.lastModified);
    }

    public final byte[] getFile(String name) throws IOException {
        return getFile(new File(name));
    }

    public final byte[] getFile(File file) throws IOException {
        if (!file.exists()) {
            return null;
        }

        FileCacheInfo finfo = (FileCacheInfo) cache.get(file);
        long lastmod = file.lastModified();
        if (finfo == null || finfo.lastModified != lastmod) {
            FileInputStream fis = new FileInputStream(file);
            byte[] content = new byte[fis.available()];
            fis.read(content);
            fis.close();
            finfo = new FileCacheInfo(file, lastmod, content);
            cache.put(file, finfo);
            return content;
        } else {
            return finfo.content;
        }
    }

    /**
     * FileCacheInfo
     */
    private class FileCacheInfo {
        public File file;
        public long lastModified = 0;
        public byte[] content;

        public FileCacheInfo(File file, long lastModified, byte[] content) {
            this.file = file;
            this.lastModified = lastModified;
            this.content = content;
        }
    }
}
