/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: ObjectPool.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.util;

/**
 * ObjectPool is an abstract Object Pool implementation.  In order to
 * create a working ObjectPool, one must extend ObjectPool and implement
 * the createObject method.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public abstract class ObjectPool {
    private ObjectQueue pool = new ObjectQueue();

    /**
     * createObject needs to be implemented in order to allow the ObjectPool
     * to produce Pooled objects.
     *
     * @return Poolable a new Poolable object instance
     */
    protected abstract Poolable createObject();

    /**
     * getObject retrieves a Poolable object from the Object pool, creating
     * a new instance if necessary.
     *
     * @return Poolable a Poolable object instance
     */
    public Poolable getObject() {
        Poolable result = (Poolable) pool.remove();
        if (result == null) {
            result = createObject();
            result.setPool(this);
        }
        return result;
    }

    /**
     * putObject should be called by Poolable objects or consumers of Poolable
     * objects to return a Pooled object to the Object Pool.
     *
     * @param object The Poolable object instance
     */
    public void putObject(Poolable object) {
        pool.add(object);
    }

    /**
     * isEmpty returns whether the ObjectPool is currently empty.
     *
     * @return Whether to ObjectPool is empty
     */
    public boolean isEmpty() {
        return pool.isEmpty();
    }
}

