/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Location.java 551434 2007-06-28 03:25:05Z vgritsenko $
 */

package org.apache.xindice.webadmin;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.Database;

/**
 * Helper class for parsing path information to get database, collection
 * and resource name.
 *
 * @version $Revision: 551434 $, $Date: 2007-06-27 23:25:05 -0400 (Wed, 27 Jun 2007) $
 */
public class Location {
    private Collection collection;
    private String name;
    private boolean isRoot;

    /**
     * Create Location instance out of path information. There are several
     * possible outcomes:
     *
     * <ul>
     * <li>Path is empty or "/" - both collection and name are null, isRoot
     * equals true;
     * <li>Path matches some collection's canonical name - collection is
     * set to that collection, name is null, isRoot equals false;
     * <li>Path matches some document's canonical name - collection is set
     * to the document's collection, name is document name, isRoot equals
     * false;
     * <li>Path could not be recognized, no database/collection match found -
     * both collection and name are null, isRoot equals false
     * </ul>
     *
     * In case there are collection and document that have the same name,
     * collection has a precedence over the document.
     * For example, if path is /db/col1/test, and collection col1 contains
     * child collection test and a document test, collection test will be
     * returned, not a document.
     *
     * @param path Path string
     * @throws DBException
     */
    public Location(String path) throws DBException {
        if (path == null || path.length() == 0 || path.equals("/")) {
            isRoot = true;
            return;
        }

        Database db = getDatabase(path);
        if (db == null) {
            collection = null;
            name = null;
            return;
        }

        StringBuffer buf = new StringBuffer(path);

        // strip database name
        String dbName = db.getName();
        buf.delete(0, dbName.length() + 1);

        Collection col;
        if (buf.length() == 0 || (buf.length() == 1 && buf.charAt(0) == '/')) {
            col = db;
        } else {
            col = db.getCollection(buf.toString());
        }

        String resourceName = null;
        if (col == null) { // no collection found -> try resource
            if (buf.charAt(buf.length() - 1) ==  '/') {
                buf.deleteCharAt(buf.length() - 1);
            }

            int split = buf.lastIndexOf("/");
            resourceName = buf.substring(split + 1, buf.length());
            String colPath = buf.substring(0, split);

            if (colPath.length() > 0) {
                col = db.getCollection(colPath);
            } else {
                col = db;
            }
        }

        collection = col;
        if (collection != null) {
            this.name = resourceName;
        }
    }

    private Database getDatabase(String path) {
        if (path.charAt(0) != '/') {
            return null;
        }

        int end = path.indexOf('/', 1);
        String dbName = (end != -1) ? path.substring(1, end) : path.substring(1);

        return Database.getDatabase(dbName);
    }

    public Collection getCollection() {
        return collection;
    }

    public String getName() {
        return name;
    }

    public boolean isRoot() {
        return isRoot;
    }
}
