/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: WebAdminManager.java 551435 2007-06-28 03:26:43Z vgritsenko $
 */

package org.apache.xindice.webadmin;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.webadmin.viewer.HtmlCollectionViewer;
import org.apache.xindice.webadmin.viewer.HtmlDatabaseViewer;
import org.apache.xindice.webadmin.viewer.HtmlResourceViewer;
import org.apache.xindice.webadmin.webdav.components.DAVComponent;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * Manager class that is responsible for reading configuration, initializing
 * all components, and dispatching calls.
 *
 * WebAdminManager configuration has several sections descibing different
 * components:
 * <ul>
 * <li>methods - list of supported WebDAV methods
 * <li>colviewers - list of collection-level operations for WebAdmin
 * <li>resviewers - list of resource-level operations for WebAdmin
 * <li>dbviewers - list of database-level operations for WebAdmin
 * </ul>
 *
 * @version $Revision: 551435 $, $Date: 2007-06-27 23:26:43 -0400 (Wed, 27 Jun 2007) $
 */
public class WebAdminManager {

    private final Map methods = new HashMap();
    private final Map colViewers = new HashMap();
    private final Map resViewers = new HashMap();
    private final Map dbViewers = new HashMap();

    private static String[] methodsList;
    private static String[] colViewersList;
    private static String[] resViewersList;
    private static String[] dbViewersList;

    private static final Log log = LogFactory.getLog(WebAdminManager.class);


    public WebAdminManager(Element config) {
        Node node = config.getElementsByTagName("methods").item(0);
        try {
            methodsList = processSection(node, "method", methods);
        } catch (Exception e) {
            log.error("Failed to load WebDAV methods", e);
        }

        node = config.getElementsByTagName("colviewers").item(0);
        try {
            colViewersList = processSection(node, "viewer", colViewers);
        } catch (Exception e) {
            log.error("Failed to load WebAdmin collection viewers", e);
        }

        node = config.getElementsByTagName("resviewers").item(0);
        try {
            resViewersList = processSection(node, "viewer", resViewers);
        } catch (Exception e) {
            log.error("Failed to load WebAdmin resource viewers", e);
        }

        node = config.getElementsByTagName("dbviewers").item(0);
        try {
            dbViewersList = processSection(node, "viewer", dbViewers);
        } catch (Exception e) {
            log.error("Failed to load WebAdmin DB viewers", e);
        }
    }

    private String[] processSection(Node node, String name, Map storage) throws Exception {
        ArrayList list = new ArrayList();

        NodeList nodes = ((Element) node).getElementsByTagName(name);
        for (int i = 0; i < nodes.getLength(); i++) {
            Node child = nodes.item(i);
            NamedNodeMap attrs = child.getAttributes();

            String className = attrs.getNamedItem("class").getNodeValue();
            Object object = Class.forName(className).newInstance();

            String id = attrs.getNamedItem("id").getNodeValue();
            if (id != null && id.length() > 0) {
                storage.put(id, object);
                list.add(id);
            }
        }

        Node attr = node.getAttributes().getNamedItem("default");
        if (attr != null) {
            storage.put("default", storage.get(attr.getNodeValue()));
        }

        return (String[]) list.toArray(new String[0]);
    }

    /**
     * Gets class that handles specific DAV request.
     *
     * @param id Method name (e.g. PROPFIND, MKCOL)
     * @return applicable class, if it has been configured, null otherwise
     */
    public DAVComponent getMethod(String id) {
        return (DAVComponent) methods.get(id);
    }

    /**
     * Gets class that handles specific collection-level operation for WebAdmin.
     *
     * @param id Opration id
     * @return applicable class, if it has been configured, null otherwise
     */
    public HtmlCollectionViewer getCollectionViewer(String id) {
        if (id == null || id.length() == 0) {
            id = "default";
        }

        return (HtmlCollectionViewer) colViewers.get(id);
    }

    /**
     * Gets class that handles specific resource-level operation for WebAdmin.
     *
     * @param id Opration id
     * @return applicable class, if it has been configured, null otherwise
     */
    public HtmlResourceViewer getResourceViewer(String id) {
        if (id == null || id.length() == 0) {
            id = "default";
        }

        return (HtmlResourceViewer) resViewers.get(id);
    }

    /**
     * Gets class that handles specific database-level operation for WebAdmin.
     *
     * @param id Opration id
     * @return applicable class, if it has been configured, null otherwise
     */
    public HtmlDatabaseViewer getDatabaseViewer(String id) {
        if (id == null || id.length() == 0) {
            id = "default";
        }

        return (HtmlDatabaseViewer) dbViewers.get(id);
    }

    /**
     * Gets all supported methods.
     *
     * @return Array of all supported WebDAV methods.
     */
    public static String[] getMethods() {
        return methodsList;
    }

    /**
     * Gets all supported database-level operations.
     *
     * @return Array of all supported database-level operations.
     */
    public static String[] getDatabaseViewers() {
        return dbViewersList;
    }

    /**
     * Gets all supported collection-level operations.
     *
     * @return Array of all supported collection-level operations.
     */
    public static String[] getCollectionViewers() {
        return colViewersList;
    }

    /**
     * Gets all supported resource-level operations.
     *
     * @return Array of all supported resource-level operations.
     */
    public static String[] getResourceViewers() {
        return resViewersList;
    }
}
