/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: UploadCollectionViewer.java 564043 2007-08-08 23:11:58Z vgritsenko $
 */

package org.apache.xindice.webadmin.viewer.components;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.webadmin.viewer.HtmlCollectionViewer;
import org.apache.xindice.webadmin.viewer.HtmlViewerComponent;
import org.apache.xindice.xml.dom.DOMParser;

import org.w3c.dom.Document;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import java.io.IOException;
import java.util.Iterator;

/**
 * Xindice Html Viewer for uploading documents to this collection.
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @version $Revision: 564043 $, $Date: 2007-08-08 19:11:58 -0400 (Wed, 08 Aug 2007) $
 */
public class UploadCollectionViewer extends HtmlViewerComponent implements HtmlCollectionViewer {

    protected static final String INLINE_METADATA = "inline-metadata";
    private static final Log log = LogFactory.getLog(UploadCollectionViewer.class);

    public void execute(HttpServletRequest req, HttpServletResponse res, Collection col)
            throws ServletException, IOException {

        String path = col.getCanonicalName();
        String title = "Upload Document to Collection " + path;

        ServletOutputStream output = startViewer(title, path, res);
        printStartTable(title, output);
        printStartSingleTableBox(output);

        String method = req.getMethod();
        if (!method.equalsIgnoreCase("POST")) {
            output.print(UPLOAD_FORM);
        } else {
            //	Create a new file upload handler
            ServletFileUpload upload = new ServletFileUpload();
            // Set upload parameters
            upload.setSizeMax(-1);
            upload.setFileItemFactory(new DiskFileItemFactory());

            Form form;
            try {
                form = parseUpload(req, upload);
            } catch (FileUploadException e) {
                log.error(e);
                throw new ServletException(e);
            }

            if (form.content == null && form.content.length == 0) {
                throw new ServletException("empty upload file");
            }

            if (form.type != 'b') { // xml or detect
                handleXML(col, form, output);
            } else {
                handleBinary(col, form, output);
            }

        }
        printEndSingleTableBox(output);
        printEndTable(output);
        finishViewer(output);
    }

    private Form parseUpload(HttpServletRequest req, ServletFileUpload upload) throws FileUploadException {
        Form form = new Form();

        // Process the uploaded fields
        Iterator i = upload.parseRequest(req).iterator();
        while (i.hasNext()) {
            FileItem item = (FileItem) i.next();
            if (!item.isFormField()) {
                if (item.getFieldName().equals("file")) {
                    form.content = item.get();
                }
            } else {
                if (item.getFieldName().equals("overwrite")) {
                    form.overwrite = (Boolean.getBoolean(item.getString()));
                } else if (item.getFieldName().equals("type")) {
                    form.type = item.getString().charAt(0);
                } else if (item.getFieldName().equals("name")) {
                    form.name = item.getString();
                }
            }
            item.delete();
        }

        return form;
    }

    private void handleXML(Collection col, Form form, ServletOutputStream out) throws IOException {
        try {
            insertXML(col, form.name, form.content, form.overwrite, out);
        } catch (XindiceException e) {
            if (form.type == 'd') {
                try {
                    insertBinary(col, form.name, form.content, form.overwrite, out);
                } catch (DBException e1) {
                    log.error(e1);
                    printError("Could not create Binary!", e1.getMessage(), out);
                }
            } else {
                printError("Not a valid xml", e.getMessage(), out);
            }
        }
    }

    private void handleBinary(Collection col, Form form, ServletOutputStream out) throws IOException {
        try {
            insertBinary(col, form.name, form.content, form.overwrite, out);
        } catch (DBException e) {
            log.error(e);
            printError("Could not create Binary!", e.getMessage(), out);
        }
    }

    protected String getName() {
        return "upload";
    }

    private void printMetaError(ServletOutputStream output) throws IOException {
        output.print("<b>Metadata storage is not enabled!<br />");
        output.print("If you want to store binary content please ");
        output.print("enable metadata storage inside the server.xml");
        output.print(" and enable Inline Metadata inside this ");
        output.print("Collection Configuration!</b>");
    }

    private void printError(String info, String message, ServletOutputStream output) throws IOException {
        output.print("<b>");
        output.print(info);
        output.print("</b><br />");
        output.print(message);
    }

    private void insertXML(Collection col, String name, byte[] content, boolean overwrite, ServletOutputStream out)
            throws XindiceException, IOException {
        Document resource = DOMParser.toDocument(content);
        if (name.trim().length() == 0) {
            name = null;
        }

        if (!overwrite) {
            try {
                name = col.insertDocument(name, resource).toString();
            } catch (DBException e) {
                if (e.faultCode == FaultCodes.COL_DUPLICATE_RESOURCE) {
                    printError("Could not create Resource!", name + " already exists!", out);
                } else {
                    throw e;
                }
            }
        } else {
            col.setDocument(name, resource);
        }
        out.print("<b>Resource <a href=\"./");
        out.print(name);
        out.print("?viewer=default\">");
        out.print(name);
        out.print("</a> created!</b>");
    }

    private void insertBinary(Collection col, String name, byte[] content, boolean overwrite, ServletOutputStream out)
            throws DBException, IOException {

        boolean inlineMetaSupported = col.getConfig().getBooleanAttribute(INLINE_METADATA, false);
        if (!inlineMetaSupported) {
            printMetaError(out);
        } else {
            if (name.trim().length() == 0) {
                name = null;
            }

            if (!overwrite) {
                try {
                    name = col.insertBinary(name, content).toString();
                } catch (DBException e) {
                    if (e.faultCode == FaultCodes.COL_DUPLICATE_RESOURCE) {
                        printError("Could not update Binary!", name + " already exists!", out);
                    } else {
                        throw e;
                    }
                }
            } else {
                col.setBinary(name, content);
            }

            out.print("<b>Binary <a href=\"./");
            out.print(name);
            out.print("?viewer=default\">");
            out.print(name);
            out.print("</a> created!</b>");
        }
    }

    private class Form {
        public byte[] content = null;
        public String name = null;
        public boolean overwrite = false;
        public char type = 'd';
    }

}
